{% extends "!base.html" %}

{% block announce %}
<p>
  Enjoying the library? Give us a
  <span class="md-icon si-icon-inline fontawesome_regular_star"></span>
  on
  <a class="reference external" href="https://github.com/mhostetter/sdr">GitHub</a>.
</p>
{% endblock %}

{% block outdated %}
You're not viewing the latest version.
<a href="{{ '../' ~ base_url}}">
  <strong>Click here to go to latest.</strong>
</a>
{% endblock %}
