"""
A subpackage for various detection algorithms.
"""

from ._approximation import *
from ._coherent_integration import *
from ._correlator import *
from ._energy import *
from ._non_coherent_integration import *
from ._theory import *
